@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2025	PgPool Global Development Group
+ * Copyright (c) 2003-2024	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -84,7 +84,7 @@ static bool connect_using_existing_connection(POOL_CONNECTION * frontend,
 static void check_restart_request(void);
 static void enable_authentication_timeout(void);
 static void disable_authentication_timeout(void);
-static int	wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr);
+static int	wait_for_new_connections(int *fds, SockAddr *saddr);
 static void check_config_reload(void);
 static void get_backends_status(unsigned int *valid_backends, unsigned int *down_backends);
 static void validate_backend_connectivity(int front_end_fd);
@@ -152,9 +152,6 @@ do_child(int *fds)
 	POOL_CONNECTION_POOL *volatile backend = NULL;
 	struct timeval now;
 	struct timezone tz;
-	struct timeval timeout;
-	static int	connected = 0;	/* non 0 if has been accepted connections from
-								 * frontend */
 
 	/* counter for child_max_connections.  "volatile" declaration is necessary
 	 * so that this is counted up even if long jump is issued due to
@@ -290,6 +287,8 @@ do_child(int *fds)
 			if (pool_config->child_max_connections > 0)
 				connections_count++;
 
+			pool_get_my_process_info()->client_connection_count++;
+
 			/* check if maximum connections count for this child reached */
 			if ((pool_config->child_max_connections > 0) &&
 				(connections_count >= pool_config->child_max_connections))
@@ -313,9 +312,6 @@ do_child(int *fds)
 	/* We can now handle ereport(ERROR) */
 	PG_exception_stack = &local_sigjmp_buf;
 
-	timeout.tv_sec = pool_config->child_life_time;
-	timeout.tv_usec = 0;
-
 	for (;;)
 	{
 		StartupPacket *sp;
@@ -337,10 +333,11 @@ do_child(int *fds)
 		/* Destroy session context for just in case... */
 		pool_session_context_destroy();
 
-		front_end_fd = wait_for_new_connections(fds, &timeout, &saddr);
+		front_end_fd = wait_for_new_connections(fds, &saddr);
+		pool_get_my_process_info()->wait_for_connect = 0;
 		if (front_end_fd == OPERATION_TIMEOUT)
 		{
-			if (pool_config->child_life_time > 0 && connected)
+			if (pool_config->child_life_time > 0 && pool_get_my_process_info()->connected)
 			{
 				ereport(DEBUG1,
 						(errmsg("child life %d seconds expired", pool_config->child_life_time)));
@@ -411,7 +408,7 @@ do_child(int *fds)
 			child_frontend = NULL;
 			continue;
 		}
-		connected = 1;
+		pool_get_my_process_info()->connected = 1;
 
 		/*
 		 * show ps status
@@ -420,6 +417,7 @@ do_child(int *fds)
 		snprintf(psbuf, sizeof(psbuf), "%s %s %s idle",
 				 sp->user, sp->database, remote_ps_data);
 		set_ps_display(psbuf, false);
+		set_process_status(IDLE);
 
 		/*
 		 * Initialize per session context
@@ -471,13 +469,12 @@ do_child(int *fds)
 		if (pool_config->log_disconnections)
 			log_disconnections(sp->database, sp->user);
 
-		timeout.tv_sec = pool_config->child_life_time;
-		timeout.tv_usec = 0;
-
 		/* increment queries counter if necessary */
 		if (pool_config->child_max_connections > 0)
 			connections_count++;
 
+		pool_get_my_process_info()->client_connection_count++;
+
 		/* check if maximum connections count for this child reached */
 		if ((pool_config->child_max_connections > 0) &&
 			(connections_count >= pool_config->child_max_connections))
@@ -1062,7 +1059,7 @@ static POOL_CONNECTION_POOL * connect_backend(StartupPacket *sp, POOL_CONNECTION
 }
 
 /*
- * signal handler for SIGTERM, SIGINT and SIGQUUT
+ * signal handler for SIGTERM, SIGINT and SIGQUIT
  */
 static RETSIGTYPE die(int sig)
 {
@@ -1135,7 +1132,7 @@ static RETSIGTYPE close_idle_connection(int sig)
 	int			main_node_id;
 
 	/*
-	 * DROP DATABSE is ongoing.
+	 * DROP DATABASE is ongoing.
 	 */
 	if (ignore_sigusr1)
 		return;
@@ -1412,7 +1409,7 @@ static void
 check_restart_request(void)
 {
 	/*
-	 * Check if restart request is set because of failback event happend.  If
+	 * Check if restart request is set because of failback event happened.  If
 	 * so, exit myself with exit code 1 to be restarted by pgpool parent.
 	 */
 	if (pool_get_my_process_info()->need_to_restart)
@@ -1429,12 +1426,12 @@ check_restart_request(void)
 /*
  * wait_for_new_connections()
  * functions calls select on sockets and wait for new client
- * to connect, on successfull connection returns the socket descriptor
- * and returns -1 if timeout has occured
+ * to connect, on successful connection returns the socket descriptor
+ * and returns -1 if timeout has occurred
  */
 
 static int
-wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
+wait_for_new_connections(int *fds, SockAddr *saddr)
 {
 	fd_set		rmask;
 	int			numfds;
@@ -1452,10 +1449,8 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 	static int	cnt;
 #endif
 
-	struct timeval *timeoutval;
-	struct timeval tv1,
-				tv2,
-				tmback = {0, 0};
+	struct timeval *timeout;
+	struct timeval timeoutdata;
 
 	for (walk = fds; *walk != -1; walk++)
 		socket_set_nonblock(*walk);
@@ -1465,24 +1460,7 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 	else
 		set_ps_display("wait for connection request", false);
 
-	memcpy((char *) &rmask, (char *) &readmask, sizeof(fd_set));
-
-	if (timeout->tv_sec == 0 && timeout->tv_usec == 0)
-		timeoutval = NULL;
-	else
-	{
-		timeoutval = timeout;
-		tmback.tv_sec = timeout->tv_sec;
-		tmback.tv_usec = timeout->tv_usec;
-		gettimeofday(&tv1, NULL);
-
-#ifdef DEBUG
-		ereport(DEBUG3,
-				(errmsg("before select = {%d, %d}", timeoutval->tv_sec, timeoutval->tv_usec)));
-		ereport(DEBUG3,
-				(errmsg("g:before select = {%d, %d}", tv1.tv_sec, tv1.tv_usec)));
-#endif
-	}
+	set_process_status(WAIT_FOR_CONNECT);
 
 	/*
 	 * If child life time is disabled and serialize_accept is on, we serialize
@@ -1539,7 +1517,41 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 				(errmsg("LOCKING select()")));
 	}
 
-	numfds = select(nsocks, &rmask, NULL, NULL, timeoutval);
+	for (;;)
+	{
+		/* check backend timer is expired */
+		if (backend_timer_expired)
+		{
+			pool_backend_timer();
+			backend_timer_expired = 0;
+		}
+
+		/* prepare select */
+		memcpy((char *) &rmask, (char *) &readmask, sizeof(fd_set));
+		if (pool_config->child_life_time > 0)
+		{
+			timeoutdata.tv_sec = 1;
+			timeoutdata.tv_usec = 0;
+			timeout = &timeoutdata;
+		}
+		else
+		{
+			timeout = NULL;
+		}
+
+		numfds = select(nsocks, &rmask, NULL, NULL, timeout);
+
+		/* not timeout*/
+		if (numfds != 0)
+			break;
+
+		/* timeout */
+		pool_get_my_process_info()->wait_for_connect++;
+
+		if (pool_get_my_process_info()->wait_for_connect > pool_config->child_life_time)
+			return OPERATION_TIMEOUT;
+
+	}
 
 	save_errno = errno;
 
@@ -1557,40 +1569,6 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 		backend_timer_expired = 0;
 	}
 
-	/*
-	 * following code fragment computes remaining timeout val in a portable
-	 * way. Linux does this automatically but other platforms do not.
-	 */
-	if (timeoutval)
-	{
-		gettimeofday(&tv2, NULL);
-
-		tmback.tv_usec -= tv2.tv_usec - tv1.tv_usec;
-		tmback.tv_sec -= tv2.tv_sec - tv1.tv_sec;
-
-		if (tmback.tv_usec < 0)
-		{
-			tmback.tv_sec--;
-			if (tmback.tv_sec < 0)
-			{
-				timeout->tv_sec = 0;
-				timeout->tv_usec = 0;
-			}
-			else
-			{
-				tmback.tv_usec += 1000000;
-				timeout->tv_sec = tmback.tv_sec;
-				timeout->tv_usec = tmback.tv_usec;
-			}
-		}
-#ifdef DEBUG
-		ereport(DEBUG3,
-				(errmsg("g:after select = {%d, %d}", tv2.tv_sec, tv2.tv_usec)));
-		ereport(DEBUG3,
-				(errmsg("after select = {%d, %d}", timeout->tv_sec, timeout->tv_usec)));
-#endif
-	}
-
 	errno = save_errno;
 
 	if (numfds == -1)
@@ -1602,12 +1580,6 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 				 errdetail("select on socket failed with error : \"%m\"")));
 	}
 
-	/* timeout */
-	if (numfds == 0)
-	{
-		return OPERATION_TIMEOUT;
-	}
-
 	for (walk = fds; *walk != -1; walk++)
 	{
 		if (FD_ISSET(*walk, &rmask))
@@ -2109,3 +2081,9 @@ static int opt_sort(const void *a, const void *b)
 {
 	return strcmp( *(char **)a, *(char **)b);
 }
+
+void
+set_process_status(ProcessStatus status)
+{
+	pool_get_my_process_info()->status = status;
+}
